// FIRCalc01.cpp : Defines the entry point for the console application.
//

#pragma warning(disable:4996)	//For VisualStudio 2005 Only

#include <math.h>
#include <stdio.h>	//for Prototyping
#include <string.h>	//for Prototyping
#include <memory.h>	//for Prototyping
#include <stdlib.h>	//for Prototyping

#define PI			3.1415926535897932384626433
#define DEFFREQ		44100
#define DEFTAPS		1023	//must be odd value. Freq/Taps = Filter Frequency accuracy.
#define DEFREDUCT	-200.0	//Default target reduction in dB.
#define DEFKISER	8.0
#define DEFLOW		0.0		//Default Low Freq
#define	DEFHIGH		40000.0	//Default High Freq

double	SampleFreq;
int		TAPS_HALF;				//Variable but treated as constant in program
double	Pass_Freq_From;			//Low Cut Freq. Frequency below this parameter is stopped.
double	Pass_Freq_To;			//High Cut Freq. Frequency over this parameter is stopped.
		// If Pass_Freq_From = 0, Pass_Freq_To = 500,    it means  Low Pass Filter, Freq = 500Hz.
		// If Pass_Freq_From = 2000, Pass_Freq_To = 100000,   it means High Pass Filter, Freq = 2000Hz.
		// If Pass_Freq_From = 300, Pass_Freq_To = 3000, it means Band Pass Filter, 300 to 3000Hz.

double EQFreqs[128];	// 128 EQ Points. Frequency.
double EQGains[128];	// 128 EQ Gains. 

double	FreqResR[DEFTAPS];			//Freq Responce Real, Desired
//double	FreqResI[DEFTAPS];			//Freq Responce Imaginal:NOT USED

double	alpha;					//Kaiser Window Constant
double	Kwindow[DEFTAPS];		//Kaiser Window Parameter
double	FIRCoeff[DEFTAPS];		//Filter Coefficient

double	Freq[(DEFTAPS-1)/2];	//Result Responce:Frequency
double	Gain[(DEFTAPS-1)/2];	//Result Responce:Amplitude
double	Theta[(DEFTAPS-1)/2];	//Result Responce:Phase


void InitCoeffParams(void);	//0 Initialize for parameters
void ReadFreqs(void);		//Read desired Frequency
void ReadEQs(void);			//Read desired Room Correction EQ values
void CreateFres(void);		//Generate desired Frequency Responce
void CalcCoeffs(void);		//Generate FIR Coeffs by IDFT
void DoKaiser(int Taps, double alpha);	//Kaiser Window Function
void OutKaiser(void);
void OutCoeffs(void);		//Output FIR Coeff Parameter to File
void CalcFres(void);		//Calculate Frequency Responce
void OutFres(void);			//Output Frequency Responce to File
double Bessel(double alpha);//0 Bessel Function

void ProcessWav(char *infilename, char *outfilename);	//Prototype Routine


int main(int argc, char* argv[])
{
	InitCoeffParams();
	ReadFreqs();
	ReadEQs();
	CreateFres();
	CalcCoeffs();
	DoKaiser(DEFTAPS, alpha);
	OutKaiser();
	OutCoeffs();
	CalcFres();
	OutFres();
	ProcessWav("C:\\Temp\\WHITENOISE.WAV","C:\\Temp\\OUTPUT.WAV");
	return 0;
}

void InitCoeffParams(void)
{
	int i;
	for(i = 0; i < DEFTAPS; i++ ) 
	{
		FIRCoeff[i] = 0.0;
	}
	SampleFreq = DEFFREQ;
	TAPS_HALF = (DEFTAPS - 1)/2;
	alpha = DEFKISER;
	Pass_Freq_From = DEFLOW;
	Pass_Freq_To = DEFHIGH;
	
	EQFreqs[0] = 0.0;
	EQGains[0] = 0.0;
	for(i = 1; i < 128; i++)
	{
		EQFreqs[i] = SampleFreq;	//Fill by Max Freq
		EQGains[i] = 0.0;
	}
}

void ReadFreqs(void)
{
	// PROTOTYPE. read from File.
	// For actual DSP, read from I2C memory / HPI etc.

	FILE * f;
	if ((f = fopen("C:\\temp\\FREQ.txt","r")) != NULL)
	{
		fscanf_s(f, "%lf,%lf",&Pass_Freq_From,&Pass_Freq_To);
		fclose(f);
	}
}

void ReadEQs(void)
{
	// PROTOTYPE. read from file.
	// For actual DSP, read from I2C memory / HPI etc.

	int i;
	FILE * f;
	if ((f = fopen("C:\\temp\\EQ.txt","r")) != NULL)
	{
		i = 1;	//Keep EQFreqs[0] = 0
		while (!feof(f))
		{
			fscanf(f, "%lf,%lf",&EQFreqs[i],&EQGains[i]);
			i += 1;
			if ( i > 126) break;
		}
		fclose(f);	
	}
}

void CreateFres(void)
{
	// gain = pow(10.0, desired_dB / 20.0);
	// There are folding at Fs/2. _____L~~~~~H_____|_____H'~~~L'____. |=Fs/2

	int i;
	double gain;

	for (i = 0; i < TAPS_HALF+1; i++)
	{
		double CurrentFreq = SampleFreq * (i + 1)/DEFTAPS;
		if (CurrentFreq < Pass_Freq_From)
		{
			//Stop Area 1
			gain = pow(10.0, DEFREDUCT / 20.0);
		}
		else if (CurrentFreq < Pass_Freq_To) 
		{
			//Pass Area 1
			//Initialize to 0 dB
			gain = pow(10.0, 0.0);

			//Find EQ freq and Value. If There are no EQ, all gains will be 0.
			for (int k = 0; k < 126; k++)
			{
				if ((EQFreqs[k] <= CurrentFreq) && (CurrentFreq < EQFreqs[k+1]))
				{
					//Linear interpolation
					gain = pow(10.0, (EQGains[k]+ (EQGains[k+1]-EQGains[k])*(CurrentFreq - EQFreqs[k])/(EQFreqs[k+1]-EQFreqs[k])) /20.0);
					break;
				}
			}
		}
		else if (CurrentFreq <= SampleFreq / 2)
		{
			//Stop Area 2
			gain = pow(10.0, DEFREDUCT / 20.0);
		}

		FreqResR[i] = gain;
		//FreqResI[i] = 0.0;		//Set 0 to Imaginary
	}
	for (i = TAPS_HALF+1; i < DEFTAPS; i++)
	{
		//Mirror Freq Responce
		FreqResR[i] = FreqResR[DEFTAPS - i - 1];
		//FreqResI[i] = FreqResI[DEFTAPS - i - 1];
	}

}


void CalcCoeffs(void)
{
	// Calculate Inverse DFT for desired Frequency Responces.
	// Result is Filter Coeffs (not windowed yet)
	int i,k;
	double omega;
	double Tempcoeff_r[DEFTAPS];	//Real,
	//double Tempcoeff_i[DEFTAPS];	//Imaginal, not used

	for (i = 0; i < DEFTAPS; i++) 
	{
		Tempcoeff_r[i] = 0.0;	//Initialize Coeffs to zero
		//Tempcoeff_i[i] = 0.0; 
		for (k = 0; k < DEFTAPS; k++) 
		{
			//Calculate Coeffs by sum up 0 to Tap Count
			omega = 2.0 * PI * k * i / DEFTAPS; 
			Tempcoeff_r[i] = Tempcoeff_r[i] + FreqResR[k] * cos(omega);	// - FreqResI[k] * sin(omega); //FreqResI is always 0.00
			//Tempcoeff_i[i] = Tempcoeff_i[i] + FreqResI[k] * cos(omega) + FreqResR[k] * sin(omega); 
		}
	}
	for (i = 0; i < DEFTAPS; i++) 
	{
		Tempcoeff_r[i] = Tempcoeff_r[i] / DEFTAPS; 
		//Tempcoeff_i[i] = Tempcoeff_i[i] / DEFTAPS;
	}
	//TempCoeff is starting from 0. should be shifted half taps.
	for (i = 0; i < TAPS_HALF; i++) 
	{
		FIRCoeff[i] = Tempcoeff_r[TAPS_HALF-i];
	}
	for (i = TAPS_HALF; i < 2*TAPS_HALF+1; i++) 
	{
		FIRCoeff[i] = Tempcoeff_r[i-TAPS_HALF];
	}
}

void DoKaiser(int Taps, double K_alpha)
{
	int i;
	double Numer, Denom;
	double center,Kg,Kd;
	
	Denom = Bessel(K_alpha);	

	center = (double)(Taps - 1)/2;
	for (i = 0; i < Taps; i++)
	{
		Kg = (i*1.0 - center) / center;		//Regulated Distance, -1 to 1
		Kd = (K_alpha * sqrt(1.0 - Kg*Kg));	//kaiser Distance
		Numer = Bessel(Kd);
		Kwindow[i] = Numer / Denom;
		//PROTOTYPE: Actually don't have to store each Kaiser Value.
	}

	for(i = 0; i < Taps; i++) 
	{
		FIRCoeff[i] = FIRCoeff[i] * Kwindow[i];
	}
}

void OutKaiser(void)
{
	//PROTOTYPE: Output to File
	FILE * f;
	f = fopen("C:\\Temp\\kaiser.txt", "w");
	for (int i = 0 ; i < DEFTAPS; i++)
	{
		fprintf(f, "%3.10lf\n",Kwindow[i]);
	}
	fclose(f);
}

double Bessel(double alpha)
{
	double delta = 1e-14;
	double BesselValue;
	double Term, k, F;
                                          
	k = 0.0;
	BesselValue = 0.0;
	F = 1.0;
	Term = 0.5;
	while( (Term < -delta) || ( delta < Term) )
	{
		k = k + 1.0;						// k = step 1,2,3,4,5
		F = F * ((alpha / 2)/k);						// f(k+1) = f(k)*(c/k),f(0)=1. c is alpha/2. c/k -> 0
		Term = F*F;							// Current term value
		BesselValue = BesselValue + Term;	// Sum(f(k)^2)
	}
	return BesselValue;
}

void OutCoeffs(void)
{
	//PROTOTYPE: Output to File
	FILE * f;
	f = fopen("C:\\Temp\\Coeffs.txt", "w");
	for (int i = 0 ; i < DEFTAPS; i++)
	{
		fprintf(f, "%3.15lf\n",FIRCoeff[i]);
	}
	fclose(f);
}
void CalcFres(void)
{

	int i, k;
	double	omega;		//angular frequency
	double	Resp_Real;	//Frequency Responce, Real
	double	Resp_Image;	//Frequency Responce, imaginary
	double	AmpABS2;	//square of amplitude characteristic
	double	Resp_Phase;	//phase characteristic

	//external variables
	//hh filter coeffs

	for(i = 0; i < TAPS_HALF; i++) 
	{
		//omega = 0 to 2Pi (rad)
		omega = (i / (DEFTAPS/360.0) * 2.0 * PI  ) / 360.0;

		//initialize
		Resp_Real = 0.00; 
		Resp_Image = 0.00;
		
		//accumulate for all TAP of FIR
		for(k =	0; k < DEFTAPS; k++) 
		{
			Resp_Real = Resp_Real + FIRCoeff[k] * cos(k * omega);
			Resp_Image = Resp_Image + FIRCoeff[k] * sin(k * omega);
		}
		//calculate square of amplitude characteristic
		AmpABS2 = Resp_Real * Resp_Real + Resp_Image * Resp_Image;

		//calculate phase characteristic
		if (Resp_Real != 0.0)
		{
			Resp_Phase = (-1) * atan(Resp_Image / Resp_Real);
		}
		else
		{
			Resp_Phase = (-1) * 0.5 * PI;
		}

		//get dB value for amplitude characteristic
		Gain[i] = 10 * log10(AmpABS2);
		Theta[i] = Resp_Phase;
		Freq[i] = SampleFreq / DEFTAPS * (i+1);
	}

}

void OutFres(void)
{
	//PROTOTYPE: Output to File
	FILE * f;
	f = fopen("C:\\Temp\\Fres.csv", "w");
	fprintf(f, "Freq[Hz],Gain[dB],Theta[rad]\n");
	for (int i = 0 ; i < TAPS_HALF; i++)
	{
		fprintf(f, "%6.1lf,%6.15lf,%10.15lf\n",Freq[i],Gain[i],Theta[i]);
	}
	fclose(f);

}
unsigned read2bytes(FILE *f)
{
    unsigned char buf[2];

    if (fread(buf, 2, 1, f) != 1) {
        fprintf(stderr, "Read error\n");
        return(1);
    }
    return 256U * buf[1] + buf[0] ;
}


unsigned long read4bytes(FILE *f)
{
    unsigned char buf[4];

    if (fread(buf, 4, 1, f) != 1) {
        fprintf(stderr, "Read error\n");
        return(1);
    }
    return ((256LU * buf[3] + buf[2]) * 256LU + buf[1] ) * 256LU + buf[0]  ;
}

void ProcessWav(char *infilename, char *outfilename)
{
    FILE *f, *fo;
    int i, channels, bits;
    unsigned long len;
    long count;
    unsigned char s[10];

	unsigned long temp_l;
	unsigned short temp_s;

	f = fopen(infilename, "rb");
	fo = fopen(outfilename, "wb");

    if (f == NULL) {
        printf("Can not open %s\n", infilename);
        return;
    }
    if (fo == NULL) {
        printf("Can not create %s\n", outfilename);
        return;
    }

    printf("finename = '%s'\n", infilename);
    if (fread(s, 4, 1, f) != 1) {
        printf("Read error\n");
        fclose(f);
        return;
    }
	
    if (memcmp(s, "RIFF", 4) != 0) {
        printf("Not a RIFF format\n");
        fclose(f);
        return;
    }

    printf("[RIFF] (%lu bytes)\n", read4bytes(f));
    if (fread(s, 8, 1, f) != 1) {
        printf("Read error\n");
        fclose(f);
        return;
    }
	
    if (memcmp(s, "WAVEfmt ", 8) != 0) {
        printf("Not a WAVEfmt format\n");
        fclose(f);
        return;
    }
    len = read4bytes(f);
    printf("[WAVEfmt ] (%lu bytes)\n", len);
    if (len != 16) {
        printf("Length of WAVEfmt must be 16\n");
        return;
    }
    printf("  Data type = %u (1 = PCM)\n", read2bytes(f));
    channels = read2bytes(f);
    printf("  Number of channels = %u (1 = mono, 2 = stereo)\n", channels);
    printf("  Sampling rate = %luHz\n", read4bytes(f));
    printf("  Bytes per second = %lu\n", read4bytes(f));
    printf("  Bytes per sample = %u\n", read2bytes(f));
    bits = read2bytes(f);
    printf("  Bits per sample = %u\n", bits);

    while (fread(s, 4, 1, f) == 1) 
	{
        len = read4bytes(f);
        s[4] = 0;
        printf("[%s] (%lu bytes)\n", s, len);
        if (memcmp(s, "data", 4) == 0) break;
        for (i = 0; i < (int)len; i++)
            printf("%02x ", fgetc(f));
        printf("\n");
    }

	//Prepare Output Wave File
	// RIFF
	s[0] = 'R';
	s[1] = 'I';
	s[2] = 'F';
	s[3] = 'F';
	fwrite(s, 1, 4, fo);
	//filesize
	temp_l = len + 36;
	fwrite(&(temp_l), sizeof(long), 1, fo);
	// WAVE
	s[0] = 'W';
	s[1] = 'A';
	s[2] = 'V';
	s[3] = 'E';
	fwrite(s, 1, 4, fo);
	// fmt chunk
	s[0] = 'f';
	s[1] = 'm';
	s[2] = 't';
	s[3] = ' ';
	fwrite(s, 1, 4, fo);
	// chunk size
	temp_l = 16;
	fwrite(&(temp_l), sizeof(long), 1, fo);
	// format PCM = 1
	temp_s = 1;
	fwrite(&(temp_s), sizeof(short), 1, fo);
	// channel stereo = 2
	temp_s = 2;
	fwrite(&(temp_s), sizeof(short), 1, fo);
	// sample 44100
	temp_l = 44100;
	fwrite(&(temp_l), sizeof(long), 1, fo);
	// Bytes per sec
	temp_l = 2 * 2 * 44100;
	fwrite(&(temp_l), sizeof(long), 1, fo);
	// Block (Bytes per sample&channel)
	temp_s = 2 * 2 ;
	fwrite(&(temp_s), sizeof(short), 1, fo);
	// Bits per sample
	temp_s = 2 * 8 ;
	fwrite(&(temp_s), sizeof(short), 1, fo);
	// data chunk
	s[0] = 'd';
	s[1] = 'a';
	s[2] = 't';
	s[3] = 'a';
	fwrite(s, 1, 4, fo);
	// data chunk size
	fwrite(&(len), sizeof(long), 1, fo);


	//get memory for "len" bytes
	unsigned char * p;
	p = (unsigned char *)malloc(len);

	//Copy File to Memory
	fread(p, len, 1,f);

	unsigned long pCounter = 0;

	double	ChannelL[DEFTAPS];		// Data Processed
	double	ChannelR[DEFTAPS];		// Data Processed
	for (i = 0; i < DEFTAPS; i++)
	{
		ChannelL[i] = 0.0;
		ChannelR[i] = 0.0;
	}

	short LData;
	short RData;
	double LDataCnv;
	double RDataCnv;

	for (count = 0; count < 0xFFFFFFFF; count++)
	{
		if (pCounter > len) break;

		//Read one data
		LData = (short)((unsigned char)(*(p+pCounter))) + (short)(((unsigned char)(*(p+pCounter+1)))<<8);
		pCounter += 2;
		RData = (short)((unsigned char)(*(p+pCounter))) + (short)(((unsigned char)(*(p+pCounter+1)))<<8);
		pCounter += 2;
		LDataCnv = ((double)LData) / 32768;
		RDataCnv = ((double)RData) / 32768;

		//Step one Data ahead. DO NOT USE SUCH A CODE FOR DSP. this code is just for understanding!
		for (int k = DEFTAPS-1; k > 0; k--)
		{
			ChannelL[k] = ChannelL[k-1];
			ChannelR[k] = ChannelR[k-1];
		}
		ChannelL[0] = LDataCnv;
		ChannelR[0] = RDataCnv;

		//Do FIR
		double Lout = 0.0;
		double Rout = 0.0;
		for (int j = 0; j < DEFTAPS; j++)
		{
			Lout = Lout + FIRCoeff[j] * ChannelL[j];
			Rout = Rout + FIRCoeff[j] * ChannelR[j];
		}
		
		// convert back to 16bit short
		LData = (short)(Lout * 32768.0);
		RData = (short)(Rout * 32768.0);
		fwrite(&(LData), sizeof(short), 1, fo);
		fwrite(&(RData), sizeof(short), 1, fo);
	}

    fclose(f);
	fclose(fo);
	free(p);
}
